﻿using System;
using System.Drawing;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing.Imaging;
using System.IO;
using SxDotNet;

namespace SxCsSdkTest
{
    public partial class Form1 : Form
    {
        const Int32 maxNumberOfCameras = 20;

        IntPtr[] camHandles = new IntPtr[maxNumberOfCameras];
        sxInfo info = new sxInfo();
        sxInfoPack[] cameraInfo = new sxInfoPack[maxNumberOfCameras];
        sxExposurePack camImgData = new sxExposurePack();
        sxExposurePack[] cameraImageData = new sxExposurePack[maxNumberOfCameras];
        sxModels modelData = new sxModels();
        Int32 numberOfCameras = 0;
        Int32 selectedCamera = 0;
        Int32 Result = 0;

        UInt16 normalisedTemperature = 2730;
        Boolean setPointChanged = true;
        Boolean imageFlag = false;
        Boolean imageMemFlag = false;
        Boolean abortExposure = false;
        static Boolean readActive = false;
        UInt32 timerValue = 0;

        Int32 pixelsInImage = 0;
        UInt16[] pixelBuffer = new UInt16[12000000]; // > largest size image currently available!
        MemoryStream ms = new MemoryStream();        // disposed of in Form Exit

        // --------------------------------------------------------
        // Constructor
        public Form1()
        {
            InitializeComponent();
        }

        // --------------------------------------------------------
        //    E V E N T S
        // --------------------------------------------------------
        // --------------------------------------------------------
        // Initialise the Form
        private void Form1_Load(object sender, EventArgs e)
        {
            buttonShutter.Enabled = false;
            buttonSetTime.Enabled = false;       // allow timer setting once camera selected
            updateTime();
            label2.Text = "DLL Version " + info.dllVersion;
            for (Int32 index = 0; index < maxNumberOfCameras; index++)
                cameraImageData[index] = camImgData;
            UInt32 usbDll = sxInfo.getUsbVersion();
            label3.Text = "USB:DLL Version " + ((usbDll >> 24) & 0xFF).ToString() + "." + ((usbDll >> 16) & 0xFF).ToString() + "." + (usbDll & 0xFF).ToString();
         }

        // --------------------------------------------------------
        // Close any open handles on exit
        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if(numberOfCameras > 0)
            {
                for (Int32 index = 0; index >= numberOfCameras; index++)
                {
                    SxSDKAccess.sxClose(camHandles[index]);
                }
            }
            ms.Dispose();   // explicitly close the memory stream
        }

        // --------------------------------------------------------
        // Connect Button
        // Find out how many cameras are fitted (maximum 20)
        // Retreive information from each of them.
        private void buttonConnect_Click(object sender, EventArgs e)
        {
            textBoxInfo.Text += "SxSDK Tester Utility\r\n";
            numberOfCameras = SxSDKAccess.sxOpen(camHandles);

            if (numberOfCameras > 0)
            {
                buttonConnect.Enabled = false;
                buttonConnect.Text = "Connected";
                textBoxInfo.Text += "Number of cameras = " + numberOfCameras.ToString() + "\r\n\r\n";

                for (Int32 index = 0; index < numberOfCameras; index++)
                {
                    textBoxInfo.Text += "Handle" + index.ToString() + " = " + camHandles[index].ToString() + "\r\n";
                    cameraInfo[index] = sxInfo.readInfoPack(camHandles[index]);//, ref tempInfo);
                    if (cameraInfo[index].result == 1)
                    {
                        // Display some of the data
                        textBoxInfo.Text += "Firmware = " + cameraInfo[index].upperFirmwareVersion.ToString() + "."
                                             + cameraInfo[index].lowerFirmwareVersion.ToString("D2") + "\r\n";
                        textBoxInfo.Text += "Build = " + cameraInfo[index].buildNumber.ToString() + "\r\n";
                        textBoxInfo.Text += "Model = 0x" + cameraInfo[index].cameraModel.ToString("X2")
                                             + ", Name = " + cameraInfo[index].cameraName + "\r\n";
                        textBoxInfo.Text += "Width = " + cameraInfo[index].imageWidth.ToString() + "\r\n";
                        textBoxInfo.Text += "Height = " + cameraInfo[index].imageHeight.ToString() + "\r\n";
                        textBoxInfo.Text += "Fitted = ";
                        if (cameraInfo[index].hasStar2K)
                            textBoxInfo.Text += "Star2K ";
                        if (cameraInfo[index].hasEEProm)
                            textBoxInfo.Text += "EEprom ";
                        if (cameraInfo[index].hasGuider)
                            textBoxInfo.Text += "Guider ";
                        if (cameraInfo[index].hasCooler)
                            textBoxInfo.Text += "Cooler ";
                        if (cameraInfo[index].hasShutter)
                            textBoxInfo.Text += "Shutter ";
                        if (cameraInfo[index].isInterlaced)
                            textBoxInfo.Text += "Interlaced CCD ";
                        textBoxInfo.Text += "\r\n\r\n";
                        comboBoxCamera.Items.Add(cameraInfo[index].cameraName);
                    }
                }
                labelSelect.Visible = true;
                comboBoxCamera.Visible = true;
            }
            else
                textBoxInfo.Text += "No Sx Cameras found!\r\n";
        }

        // --------------------------------------------------------
        // Selection of a connected camera
        private void comboBoxCamera_SelectedIndexChanged(object sender, EventArgs e)
        {
            buttonSetTime.Enabled = true;       // allow timer setting once camera selected
            timer1.Enabled = false;             // assume no cooler
            checkBoxCooler.Checked = false;     // assume no cooler
            checkBoxCooler.Enabled = false;     // assume no cooler
            buttonShutter.Enabled = false;      // assume no shutter
            selectedCamera = comboBoxCamera.SelectedIndex;
            textBoxInfo.Text += "selectedCamera = " + selectedCamera.ToString() + "\r\n";
            textBoxInfo.Text += "hasShutter = " + cameraInfo[selectedCamera].hasShutter.ToString() + "\r\n";
            textBoxInfo.Text += "Name = " + cameraInfo[selectedCamera].cameraName + "\r\n";
            if (cameraInfo[selectedCamera].hasShutter)
            {
                buttonShutter.Enabled = true;
                sxUtilities.closeShutter(camHandles[selectedCamera]);
            }
            if (cameraInfo[selectedCamera].hasCooler)
            {
                timer1.Enabled = true;
                checkBoxCooler.Enabled = true;
                checkBoxCooler.Checked = true;
                setPointChanged = true;
            }
        }

        // --------------------------------------------------------
        // Use to control the shutter if the camera is equipped with one.
        private void buttonShutter_Click(object sender, EventArgs e)
        {
            Int32 returnData = 0;

            if (buttonShutter.Text == "Close")
            {
                returnData = sxUtilities.closeShutter(camHandles[selectedCamera]);
                buttonShutter.Text = "Open";
            }
            else
            {
                returnData = sxUtilities.openShutter(camHandles[selectedCamera]);
                buttonShutter.Text = "Close";
            }
            textBoxInfo.Text += "Shutter Data = 0x" + returnData.ToString("X2") + "\r\n";
        }

        // --------------------------------------------------------
        // Sets the Cooler temperature
        private void trackBarTemp_ValueChanged(object sender, EventArgs e)
        {
            normalisedTemperature = (UInt16)(trackBarTemp.Value + 2730);
            labelSetTemp.Text = ((Single)((normalisedTemperature - 2730.0) / 10.0)).ToString();
            setPointChanged = true;
        }

        // --------------------------------------------------------
        // Cooler ON/OFF (only enabled if if fitted)
        private void checkBoxCooler_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxCooler.Checked)
            {
                checkBoxCooler.Text = "ON";
                setPointChanged = true;         // ensure set point is updated after switch on
            }
            else
                checkBoxCooler.Text = "OFF";
        }

        // --------------------------------------------------------
        // Use timer to update cooler temperature
        // Will only update if camera is fitted with cooler and there
        // is no image download pending or in progress
        private void timer1_Tick(object sender, EventArgs e)
        {
            if (!readActive && cameraInfo[selectedCamera].hasCooler)
            {
                if (setPointChanged)
                {
                    setPointChanged = false;
                    sxCoolerPack coolerInfo = new sxCoolerPack();
                    coolerInfo.camHandle = camHandles[selectedCamera];
                    coolerInfo.coolerON = checkBoxCooler.Checked;
                    coolerInfo.setPoint = (Single)((normalisedTemperature - 2730.0) / 10.0);

                    sxUtilities.cooler(coolerInfo);
                    Boolean actualState = coolerInfo.coolerState;
                    labelActualTemp.Text = coolerInfo.currentTemperature.ToString();
                }
                else
                    labelActualTemp.Text = sxUtilities.coolerTemp(camHandles[selectedCamera]).ToString();

            }
        }

        // --------------------------------------------------------
        // Displays the image on a panel converting from pixel array to 8 bit BMP
        private void panelImage_Paint(object sender, PaintEventArgs e)
        {
            if (imageFlag)
            {
                if (!imageMemFlag)
                {
                    Bitmap imageSxCam = new Bitmap(cameraInfo[selectedCamera].imageWidth, cameraInfo[selectedCamera].imageHeight, PixelFormat.Format8bppIndexed);
                    BitmapData bmpData = imageSxCam.LockBits(new Rectangle(0, 0, imageSxCam.Width, imageSxCam.Height), ImageLockMode.WriteOnly, imageSxCam.PixelFormat);
                    Int32 numBytes = bmpData.Stride * imageSxCam.Height;
                    Int32 i, j, k, index;

                    //Construct the Grayscale colour Palette
                    ColorPalette grayScalePalette = imageSxCam.Palette;
                    for (i = 0; i < 256; i++)
                        grayScalePalette.Entries[i] = Color.FromArgb(255, i, i, i);
                    imageSxCam.Palette = grayScalePalette;

                    // Take msbs of pixels (Int16) array & store in byte array for bmp data
                    Byte[] bmpBuffer = new Byte[numBytes];
                    index = 0;      // index to Int16 pixels in pixelBuffer
                    for (i = 0; i < imageSxCam.Height; i++)
                    {
                        k = i * bmpData.Stride;
                        for (j = 0; j < imageSxCam.Width; j++)
                            bmpBuffer[k++] = (Byte)(pixelBuffer[index++] >> 8);
                    }

                    //Copy the data from the byte array into BitmapData.Scan0
                    Marshal.Copy(bmpBuffer, 0, bmpData.Scan0, numBytes);

                    //Unlock the pixels
                    imageSxCam.UnlockBits(bmpData);

                    // Create Point for upper-left corner of image.
                    Point ulCorner = new Point(0, 0);

                    // Draw image to screen.
                    e.Graphics.DrawImage(imageSxCam, ulCorner);
                    imageSxCam.Dispose();
                    imageFlag = false;
                }
                else
                {
                    Bitmap imageSxCam = new Bitmap(ms);
                    imageMemFlag = false;

                    // Create Point for upper-left corner of image.
                    Point ulCorner = new Point(0, 0);

                    // Draw image to screen.
                    e.Graphics.DrawImage(imageSxCam, ulCorner);
                    imageSxCam.Dispose();
                    imageFlag = false;
                }

            }
        }

        // --------------------------------------------------------
        // Takes an Image from the selected camera. Returns when the image is downloaded
        private void buttontakeImage_Click(object sender, EventArgs e)
        {
            takeImage(false);
        }

        // --------------------------------------------------------
        // Exposes the image only - must be follwed by 
        private void buttonExpRead_Click(object sender, EventArgs e)
        {
            {
                Boolean waitForExposureEnd = false;
                pixelsInImage = cameraInfo[selectedCamera].imageWidth * cameraInfo[selectedCamera].imageHeight;
                labelPixelNumber.Text = "0";
                if (cameraInfo[selectedCamera].hasShutter)
                    sxUtilities.closeShutter(camHandles[selectedCamera]);
                cameraImageData[selectedCamera].camHandle = camHandles[selectedCamera];
                cameraImageData[selectedCamera].flags = SxSDKAccess.CCD_FLAGS_FIELD_BOTH; // specify both fields for H series cameras
                cameraImageData[selectedCamera].useGuider = false;
                cameraImageData[selectedCamera].useGated = false;
                cameraImageData[selectedCamera].xOffset = 0;
                cameraImageData[selectedCamera].yOffset = 0;
                cameraImageData[selectedCamera].imageWidth = (UInt16)cameraInfo[selectedCamera].imageWidth;
                cameraImageData[selectedCamera].imageHeight = (UInt16)cameraInfo[selectedCamera].imageHeight;
                cameraImageData[selectedCamera].xBin = 1;
                cameraImageData[selectedCamera].yBin = 1;
                cameraImageData[selectedCamera].exposureTime = timerValue;
                // For exposures of less than 1 second wait for the exposure to complete
                if (timerValue < 1000)
                    waitForExposureEnd = true;

                if (cameraInfo[selectedCamera].hasShutter)
                    sxUtilities.openShutter(camHandles[selectedCamera]);
                Result = sxImage.makeExposure(cameraImageData[selectedCamera], waitForExposureEnd);
                textBoxInfo.Text += "Result = " + Result.ToString() + ",  Time remaining = " + sxUtilities.getTimer(cameraImageData[selectedCamera].camHandle).ToString() + "\r\n";

                // Start the exposure checking timer
                if (!waitForExposureEnd)
                {
                    timerExp.Enabled = true;
                    buttonAbort.Enabled = true;
                }
                else        // Short exposure is already completed so read image now
                {
                    if (cameraInfo[selectedCamera].hasShutter)
                        sxUtilities.closeShutter(camHandles[selectedCamera]);
                    getImage(0);
                }
            }
        }

        // --------------------------------------------------------
        // This timer is used to monitor the exposure progress,
        // which allows the exposure to be aborted if needed.
        // If the exposure has completed then the image is downloaded.
        // During the image download the "readActive" flag is set
        // This flag should be used to stop any camera communications
        // which requires returned data, e.g. requesting the cooler temperature.
        private void timerExp_Tick(object sender, EventArgs e)
        {
            if (abortExposure == true)
            {
                timerExp.Enabled = false;
                buttonAbort.Enabled = false;
                abortExposure = false;
                labelGetTime.Text = "Time remaining = 0 ms";
                textBoxInfo.Text += "Exposure Aborted" + "\r\n";
            }

            // Get the time left, if zero then exposure completed.
            UInt32 currentTime = sxUtilities.getTimer(cameraImageData[selectedCamera].camHandle);
            if (currentTime == 0)
            {
                timerExp.Enabled = false;
                buttonAbort.Enabled = false;
                abortExposure = false;
                labelGetTime.Text = "Time remaining = 0 ms";
                if (cameraInfo[selectedCamera].hasShutter)
                    sxUtilities.closeShutter(camHandles[selectedCamera]);
                getImage(0);
            }
            else
            {
                labelGetTime.Text = "Time remaining = " + currentTime.ToString() + "ms";
            }

        }

        // --------------------------------------------------------
        // When the exposure Abort button is pressed set the "abortExposure" flag
        private void buttonAbort_Click(object sender, EventArgs e)
        {
            abortExposure = true;
        }

        // --------------------------------------------------------
        // Gated exposure only uses the blocking "takeImage" method
        private void buttonGated_Click(object sender, EventArgs e)
        {
            takeImage(true);
        }

        // --------------------------------------------------------
        // Monitors the camera timer when active (test program)
        private void timer2_Tick(object sender, EventArgs e)
        {
            UInt32 time = sxUtilities.getTimer(camHandles[selectedCamera]);
            if (time == 0)
            {
                timer2.Enabled = false;
                labelGetTime.Text = "Time remaining = 0 ms";
            }
            else
                labelGetTime.Text = "Time remaining = " + time.ToString() + "ms";
        }
        // --------------------------------------------------------
        private void radioButtonMinutes_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        // --------------------------------------------------------
        private void radioButtonSeconds_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        // --------------------------------------------------------
        private void radioButtonTenths_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        // --------------------------------------------------------
        private void radioButtonMs_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        // --------------------------------------------------------
        // Sets exposure time along with  time range Radio buttons
        private void trackBarExposure_ValueChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        // --------------------------------------------------------
        // Sets the camera timer (test program)
        private void buttonSetTime_Click(object sender, EventArgs e)
        {
            if ((numberOfCameras > 0) && (timerValue > 0))    // must be connected
            {
                Int32 Result = sxUtilities.setTimer(camHandles[selectedCamera], timerValue);
                timer2.Enabled = true;
            }
        }


        // --------------------------------------------------------
        //    P R I V A T E  M E T H O D S
        // --------------------------------------------------------
        // --------------------------------------------------------
        // If exposure time > 0 then uses the makeImage method
        // else it reads the image data straight away
        private void getImage(Int32 expTime)
        {
            Int32 time;

            readActive = true;
            time = Environment.TickCount & Int32.MaxValue;
            if (expTime == 0)
                Result = sxImage.readImage(cameraImageData[selectedCamera], ref pixelBuffer);
            else
                Result = sxImage.makeImage(cameraImageData[selectedCamera], ref pixelBuffer);

            time = (Environment.TickCount & Int32.MaxValue) - time;
            time -= expTime;
            labelDn.Text = time.ToString() + " ms   ";
            readActive = false;
            textBoxInfo.Text += "Read result = " + Result.ToString() + "\r\n";
            labelPixelNumber.Text = pixelsInImage.ToString();
            imageFlag = true;
            panelImage.Invalidate();
        }

        // --------------------------------------------------------
        // Prepares the sxExposurePack structure with the exposure data
        // & initiates the exposure
        private void takeImage(Boolean gated)
        {
            pixelsInImage = cameraInfo[selectedCamera].imageWidth * cameraInfo[selectedCamera].imageHeight;
            labelPixelNumber.Text = "0";

            cameraImageData[selectedCamera].camHandle = camHandles[selectedCamera];
            cameraImageData[selectedCamera].flags = SxSDKAccess.CCD_FLAGS_FIELD_BOTH; // specify both fields for H series cameras
            cameraImageData[selectedCamera].useGuider = false;
            cameraImageData[selectedCamera].useGated = gated;
            cameraImageData[selectedCamera].xOffset = 0;
            cameraImageData[selectedCamera].yOffset = 0;
            cameraImageData[selectedCamera].imageWidth = (UInt16)cameraInfo[selectedCamera].imageWidth;
            cameraImageData[selectedCamera].imageHeight = (UInt16)cameraInfo[selectedCamera].imageHeight;
            cameraImageData[selectedCamera].xBin = 1;
            cameraImageData[selectedCamera].yBin = 1;
            cameraImageData[selectedCamera].exposureTime = timerValue;

            getImage((Int32)timerValue);
            labelPixelNumber.Text = pixelsInImage.ToString();
        }

        // --------------------------------------------------------
        // Sets the exposure time in timerValue in units of ms using the track bar & radio buttons
        private void updateTime()
        {
            Int32 multiplier = 1000;        // assume milliseconds
            if (radioButtonMinutes.Checked)
            {
                multiplier = 60000;
                labelSetTime.Text = trackBarExposure.Value.ToString() + " mins";
            }
            if (radioButtonSeconds.Checked)
            {
                multiplier = 1000;
                labelSetTime.Text = trackBarExposure.Value.ToString() + " secs";
            }
            if (radioButtonTenths.Checked)
            {
                multiplier = 100;
                labelSetTime.Text = (trackBarExposure.Value * 100).ToString() + " ms";
            }
            if (radioButtonMs.Checked)
            {
                multiplier = 1;
                labelSetTime.Text = (trackBarExposure.Value).ToString() + " ms";
            }
            timerValue = (UInt32)(trackBarExposure.Value * multiplier);
        }



    }
}
