﻿using System;
using System.Runtime.InteropServices;
using System.Reflection;

namespace SxDotNet
{
    public class sxInfo
    {
        public String dllVersion
        {
            get
            {
                string appName = Assembly.GetAssembly(this.GetType()).Location;
                AssemblyName assemblyName = AssemblyName.GetAssemblyName(appName);
                return (assemblyName.Version.ToString());
            }
        }

        // ------------------------------------------------------------
        // Returns the sxUSB.dll version, packed into a UInt32
        // bits 31-24 undefined, bits 23-16 = MSD, bits 15-8 = LSD, bits 7-0 = BUILD
        public static UInt32 getUsbVersion()
        {
            return (SxSDKAccess.sxGetDLLVersion());
        }
                
        // ------------------------------------------------------------
        // Handles pointer & unmanaged memory for "readInfoPack()" 
        public static sxInfoPack readInfoPack(IntPtr camHandle)
        {
            t_sxccd_params camInfo = new t_sxccd_params();
            sxInfoPack camData = new sxInfoPack();

            // Initialize unmanged memory to hold the struct.
            IntPtr camInfoPtr = Marshal.AllocHGlobal(Marshal.SizeOf(camInfo));
            try
            {
                camData.result = (Int32)SxSDKAccess.sxGetCameraParams(camHandle, 0, camInfoPtr);
                // Copy the struct in unmanaged memory back to managed memory
                camInfo = (t_sxccd_params)Marshal.PtrToStructure(camInfoPtr, typeof(t_sxccd_params));
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(camInfoPtr);
            }

            // Transfer data from camInfo structure to camData object
            camData.imageWidth         = (Int32)camInfo.width;
            camData.imageHeight        = (Int32)camInfo.height;
            camData.pixelHeight        = camInfo.pix_height;
            camData.pixelWidth         = camInfo.pix_width;
            camData.colorMatrix        = (Int32)camInfo.color_matrix;
            camData.bitsPerPixel       = camInfo.bits_per_pixel;
            camData.blackPixelsAtStart = (Int32)camInfo.hfront_porch;
            camData.blackPixelsAtEnd   = (Int32)camInfo.hback_porch;
            camData.blackLinesAtStart  = (Int32)camInfo.vfront_porch;
            camData.blackLinesAtEnd    = (Int32)camInfo.vback_porch;
            // Star2K
            if ((camInfo.extra_caps & 0x1) > 0)
                camData.hasStar2K = true;
            else
                camData.hasStar2K = false;
            // EEprom
            if ((camInfo.extra_caps & 0x4) > 0)
                camData.hasEEProm = true;
            else
                camData.hasEEProm = false;
            // Guider
            if ((camInfo.extra_caps & 0x8) > 0)
                camData.hasGuider = true;
            else
                camData.hasGuider = false;
            // Cooler
            if ((camInfo.extra_caps & 0x10) > 0)
                camData.hasCooler = true;
            else
                camData.hasCooler = false;
            // Shutter
            if ((camInfo.extra_caps & 0x20) > 0)
                camData.hasShutter = true;
            else
                camData.hasShutter = false;
            // Camera Model
            camData.cameraModel = (Int32)SxSDKAccess.sxGetCameraModel(camHandle);
            // Camera name
            camData.cameraName = "Unknown";
            switch (camData.cameraModel)
            {
                case sxModels.MX5:  camData.cameraName = sxModels.nameMX5; break;
                case sxModels.MX5C: camData.cameraName = sxModels.nameMX5C; break;
                case sxModels.MX7:  camData.cameraName = sxModels.nameMX7; break;
                case sxModels.MX7C: camData.cameraName = sxModels.nameMX7C; break;
                case sxModels.MX8C: camData.cameraName = sxModels.nameMX8C; break;
                case sxModels.MX9:  camData.cameraName = sxModels.nameMX9; break;
                case sxModels.M25C: camData.cameraName = sxModels.nameM25C; break;
                case sxModels.M26C: camData.cameraName = sxModels.nameM26C; break;
                case sxModels.H5:   camData.cameraName = sxModels.nameH5; break;
                case sxModels.H5C:  camData.cameraName = sxModels.nameH5C; break;
                case sxModels.H9:   camData.cameraName = sxModels.nameH9; break;
                case sxModels.H9C:  camData.cameraName = sxModels.nameH9C; break;
                case sxModels.H16:  camData.cameraName = sxModels.nameH16; break;
                case sxModels.H18:  camData.cameraName = sxModels.nameH18; break;
                case sxModels.H35:  camData.cameraName = sxModels.nameH35; break;
                case sxModels.H36:  camData.cameraName = sxModels.nameH36; break;
                case sxModels.H290: camData.cameraName = sxModels.nameH290; break;
                case sxModels.H694: camData.cameraName = sxModels.nameH694; break;
                case sxModels.H674: camData.cameraName = sxModels.nameH674; break;
                case sxModels.H694C: camData.cameraName = sxModels.nameH694C; break;
                case sxModels.H674C: camData.cameraName = sxModels.nameH674C; break;
                case sxModels.SS9:  camData.cameraName = sxModels.nameSS9; break;
                case sxModels.SS9C: camData.cameraName = sxModels.nameSS9C; break;
                case sxModels.COST: camData.cameraName = sxModels.nameCOST; break;
                case sxModels.BR694: camData.cameraName = sxModels.nameBR694; break;
                case sxModels.LX1:  camData.cameraName = sxModels.nameLX1; break;
                case sxModels.H825: camData.cameraName = sxModels.nameH825; break;
                case sxModels.H825C: camData.cameraName = sxModels.nameH825C; break;
                case sxModels.USM: camData.cameraName = sxModels.nameUltraStarM; break;
                case sxModels.USC: camData.cameraName = sxModels.nameUltraStarC; break;
            }

            // Interlaced CCD
            if ((((camData.cameraModel & sxModels.MxBit) > 0) || (camData.cameraModel == sxModels.MX5C)) && (camData.cameraModel != sxModels.M25C) && (camData.cameraModel != sxModels.LX1))
                camData.isInterlaced = true;
            else
                camData.isInterlaced = false;

            // Firmware version & build number
            camData.buildNumber = 0;
            Int32 Firmware = (Int32)SxSDKAccess.sxGetFirmwareVersion(camHandle);
            camData.upperFirmwareVersion = Firmware >> 16;
            camData.lowerFirmwareVersion = Firmware & 0xFFFF;
            if (!((camData.upperFirmwareVersion < 1) || (camData.lowerFirmwareVersion < 16)))
                camData.buildNumber = (Int32)SxSDKAccess.sxGetBuildNumber(camHandle);

            return (camData);
        }
    }

    // ------------------------------------------------------------
    // Reads an Image from the camera
    public class sxImage
    {
        // Handles pointer & unmanaged memory for "sxReadPixels()" 
        public static Int32 getPixels(IntPtr camHandle, Int32 numberOfPixels, ref UInt16[] pixelBuffer)
        {
            Int32 Result = 0;

            // Initialize unmanaged memory to hold the pixel buffer.
            Int16[] pxBuffer = new Int16[numberOfPixels];
            Int32 bufferSize = Marshal.SizeOf(pxBuffer[0]) * numberOfPixels;
            IntPtr pixelBufferPtr = Marshal.AllocHGlobal(bufferSize);
            try
            {
                Result = SxSDKAccess.sxReadPixels(camHandle, pixelBufferPtr, (UInt32)numberOfPixels);
                // Copy the pixelBuffer in unmanaged memory back to managed memory
                Marshal.Copy(pixelBufferPtr, pxBuffer, 0, numberOfPixels);
                System.Buffer.BlockCopy(pxBuffer, 0, pixelBuffer, 0, numberOfPixels * 2);
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(pixelBufferPtr);
            }
            return (Result);
        }

        // ------------------------------------------------------------
        // Takes exposure & reads pixels from camera when exposure complete
        // using information in sxExposurePack
        // One stop Image operation although it blocks for the whole exposure time as well
        // as the image download period.
        // For more flexibility use makeExposure & readImage on long exposures (> 1 second)
        // Use this for short exposures (1 second or less).
        public static Int32 makeImage(sxExposurePack exposureInfo, ref UInt16[] pixelBuffer)
        {
            Int32 Result = 0;
            Int32 pixelsInImage = (exposureInfo.imageWidth / exposureInfo.xBin) * (exposureInfo.imageHeight / exposureInfo.yBin);
            Int32 pixelA = 0, pixelB = 0, line = 0, linePixel = 0, finishedWidth = 0, finishedHeight = 0, count = 0;
            UInt16 cameraModel = (UInt16)SxSDKAccess.sxGetCameraModel(exposureInfo.camHandle);
            //  UInt16 scale = (UInt16)((cameraModel == sxModels.M25C) ? 2 : 1);		 	 // Adjust width & Height for M25
            UInt16 scale = 1;
            UInt16 actualWidth = (UInt16)(exposureInfo.imageWidth * scale);
            UInt16 actualHeight = (UInt16)(exposureInfo.imageHeight / scale);
            UInt16 actualX = (UInt16)(exposureInfo.xOffset * scale);
            UInt16 actualY = (UInt16)(exposureInfo.yOffset / scale);
            finishedHeight = actualHeight * 2;
            finishedWidth = actualWidth / 2;

            Result = SxSDKAccess.sxClearPixels(exposureInfo.camHandle, SxSDKAccess.CCD_WIPE_ALL, (UInt16)(exposureInfo.useGuider ? 1 : 0));
            if (!exposureInfo.useGated)
                Result = SxSDKAccess.sxExposePixels(exposureInfo.camHandle, exposureInfo.flags, (UInt16)(exposureInfo.useGuider ? 1 : 0),
                                                      actualX, actualY, actualWidth, actualHeight,
                                                      exposureInfo.xBin, exposureInfo.yBin, exposureInfo.exposureTime);
            else
                Result = SxSDKAccess.sxExposePixelsGated(exposureInfo.camHandle, exposureInfo.flags, (UInt16)(exposureInfo.useGuider ? 1 : 0),
                                                      actualX, actualY, actualWidth, actualHeight,
                                                      exposureInfo.xBin, exposureInfo.yBin, exposureInfo.exposureTime);

            Int16[] pxBuffer = new Int16[pixelsInImage];
            Int32 bufferSize = Marshal.SizeOf(pxBuffer[0]) * pixelsInImage;
            IntPtr pixelBufferPtr = Marshal.AllocHGlobal(bufferSize);
            try
            {
                Result = SxSDKAccess.sxReadPixels(exposureInfo.camHandle, pixelBufferPtr, (UInt32)pixelsInImage);
                // Copy the pixelBuffer in unmanaged memory back to managed memory
                Marshal.Copy(pixelBufferPtr, pxBuffer, 0, pixelsInImage);
                System.Buffer.BlockCopy(pxBuffer, 0, pixelBuffer, 0, pixelsInImage * 2);

                // Adjust the Readout for M25C
                if (cameraModel == sxModels.M25C)
                {
                    UInt16[] tempBuffer = new UInt16[pixelsInImage];
                    for (count = 0; count < pixelsInImage; count++)
                        tempBuffer[count] = pixelBuffer[count];

                    for (line = 0; line < actualHeight; line++)
                    {
                        pixelA = line * actualWidth;
                        pixelB = pixelA;
                        for (linePixel = 0; linePixel < finishedWidth; linePixel++)
                        {
                            pixelBuffer[pixelA] = tempBuffer[pixelB];
                            pixelB++;
                            pixelBuffer[pixelA + finishedWidth] = tempBuffer[pixelB];
                            pixelB++;
                            pixelA++;
                        }
                    }
                }
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(pixelBufferPtr);
            }
            return (Result);
        }

        // ------------------------------------------------------------
        // Use this directly after "makeExposure"
        // Reads exposed image immmediately using information in sxExposurePack
        // The exposure time data is ignored
        public static Int32 readImage(sxExposurePack exposureInfo, ref UInt16[] pixelBuffer)
        {
            Int32 Result = 0;
            Int32 pixelsInImage = (exposureInfo.imageWidth / exposureInfo.xBin) * (exposureInfo.imageHeight / exposureInfo.yBin);
            Int32 pixelA = 0, pixelB = 0, line = 0, linePixel = 0, finishedWidth = 0, finishedHeight = 0, count = 0;
            UInt16 cameraModel = (UInt16)SxSDKAccess.sxGetCameraModel(exposureInfo.camHandle);
            //  UInt16 scale = (UInt16)((cameraModel == 0x59) ? 2 : 1);		 	 // Adjust width & Height for M25
            UInt16 scale = 1;
            UInt16 actualWidth = (UInt16)(exposureInfo.imageWidth * scale);
            UInt16 actualHeight = (UInt16)(exposureInfo.imageHeight / scale);
            UInt16 actualX = (UInt16)(exposureInfo.xOffset * scale);
            UInt16 actualY = (UInt16)(exposureInfo.yOffset / scale);
            finishedHeight = actualHeight * 2;
            finishedWidth = actualWidth / 2;

            Result = SxSDKAccess.sxClearPixels(exposureInfo.camHandle, SxSDKAccess.CCD_FLAGS_NOWIPE_FRAME, (UInt16)(exposureInfo.useGuider ? 1 : 0));
            Result = SxSDKAccess.sxLatchPixels(exposureInfo.camHandle, exposureInfo.flags, (UInt16)(exposureInfo.useGuider ? 1 : 0),
                                                      actualX, actualY, actualWidth, actualHeight,
                                                      exposureInfo.xBin, exposureInfo.yBin);

            Int16[] pxBuffer = new Int16[pixelsInImage];
            Int32 bufferSize = Marshal.SizeOf(pxBuffer[0]) * pixelsInImage;
            IntPtr pixelBufferPtr = Marshal.AllocHGlobal(bufferSize);
            try
            {
                Result = SxSDKAccess.sxReadPixels(exposureInfo.camHandle, pixelBufferPtr, (UInt32)pixelsInImage);
                // Copy the pixelBuffer in unmanaged memory back to managed memory
                Marshal.Copy(pixelBufferPtr, pxBuffer, 0, pixelsInImage);
                System.Buffer.BlockCopy(pxBuffer, 0, pixelBuffer, 0, pixelsInImage * 2);

                // Adjust the Readout for M25C
                if (cameraModel == sxModels.M25C)
                {
                    UInt16[] tempBuffer = new UInt16[pixelsInImage];
                    for (count = 0; count < pixelsInImage; count++)
                        tempBuffer[count] = pixelBuffer[count];

                    for (line = 0; line < actualHeight; line++)
                    {
                        pixelA = line * actualWidth;
                        pixelB = pixelA;
                        for (linePixel = 0; linePixel < finishedWidth; linePixel++)
                        {
                            pixelBuffer[pixelA] = tempBuffer[pixelB];
                            pixelB++;
                            pixelBuffer[pixelA + finishedWidth] = tempBuffer[pixelB];
                            pixelB++;
                            pixelA++;
                        }
                    }
                }
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(pixelBufferPtr);
            }
            return (Result);
        }

        // ------------------------------------------------------------
        // Takes exposure using information in sxExposurePack
        // Use "readImage" directly after your exposure is complete
        public static Int32 makeExposure(sxExposurePack exposureInfo, Boolean waitForExposureCompleted)
        {
            Int32 Result = 0;

            Result = SxSDKAccess.sxClearPixels(exposureInfo.camHandle, SxSDKAccess.CCD_WIPE_ALL, (UInt16)(exposureInfo.useGuider ? 1 : 0));
            sxUtilities.setTimer(exposureInfo.camHandle, exposureInfo.exposureTime);

            if (waitForExposureCompleted)
            {
                while (sxUtilities.getTimer(exposureInfo.camHandle) > 0) ;
            }

            return (Result);
        }
    }

    // ----------------------------------------------------------------
    public class sxSerialPort
    {
        // ------------------------------------------------------------
        // Handles pointer & unmanaged memory for "sxReadSerial()" 
        public static UInt32 receiveSerialData(IntPtr camHandle, Int32 camIndex, UInt16 portIndex, UInt16 numberOfChars, ref Byte[] serialBuffer)
        {
            UInt32 Result = 0;

            // Initialize unmanged memory to hold the serial buffer.
            Int32 bufferSize = Marshal.SizeOf(serialBuffer[0]) * serialBuffer.Length;
            IntPtr serialBufferPtr = Marshal.AllocHGlobal(bufferSize);
            try
            {
                Result = SxSDKAccess.sxReadSerialPort(camHandle, portIndex, numberOfChars, serialBufferPtr);
                // Copy received data in serialBuffer in unmanaged memory to managed memory
                Marshal.Copy(serialBufferPtr, serialBuffer, 0, serialBuffer.Length);
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(serialBufferPtr);
            }
            return (Result);
        }

        // ------------------------------------------------------------
        // Handles pointer & unmanaged memory for "sxWriteSerial()" 
        public static UInt32 sendSerialData(IntPtr camHandle, Int32 camIndex, UInt16 portIndex, UInt16 flush, UInt16 numberOfChars, ref Byte[] sendBuffer)
        {
            UInt32 Result = 0;

            // Initialize unmanged memory to hold the serial buffer.
            Int32 bufferSize = Marshal.SizeOf(sendBuffer[0]) * sendBuffer.Length;
            IntPtr serialBufferPtr = Marshal.AllocHGlobal(bufferSize);
            try
            {
                // Copy the sendBuffer in managed memory to unmanaged memory
                Marshal.Copy(sendBuffer, 0, serialBufferPtr, numberOfChars);
                Result = SxSDKAccess.sxWriteSerialPort(camHandle, portIndex, flush, numberOfChars, serialBufferPtr);
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(serialBufferPtr);
            }
            return (Result);
        }
    }

    // ----------------------------------------------------------------
    public class sxUtilities
    {
        // ------------------------------------------------------------
        // Reset the Camera
        public static void resetCamera(IntPtr camHandle)
        {
            SxSDKAccess.sxReset(camHandle);
        }

        // ------------------------------------------------------------
        // Sets the Star2000 Guider port
        public static Int32 setGuiderPort(IntPtr camHandle, Byte portValue)
        {
            return((Int32)SxSDKAccess.sxSetSTAR2000(camHandle, portValue));
        }

        // ------------------------------------------------------------
        // Opens the shutter
        public static Int32 openShutter(IntPtr camHandle)
        {
            return ((Int32)SxSDKAccess.sxSetShutter(camHandle, 0));
        }

        // ------------------------------------------------------------
        // Closes the shutter
        public static Int32 closeShutter(IntPtr camHandle)
        {
            return ((Int32)SxSDKAccess.sxSetShutter(camHandle, 1));
        }

        // ------------------------------------------------------------
        // Controls the cooler
        public static sxCoolerPack cooler(sxCoolerPack coolerInfo)
        {
            Int16 coolerData = 0;   // returned temperature
            UInt16 setPoint = (UInt16)((coolerInfo.setPoint + 273.0) * 10.0); // cooler set point as integer
            Byte  coolerStatus = 0; // assume OFF
            if(coolerInfo.coolerON)
                coolerStatus = 1;

            // Initialize unmanaged memory to hold the struct.
            IntPtr coolerDataPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerData));
            IntPtr coolerStatusPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerStatus));
            try
            {
                coolerInfo.result = (Int32)SxSDKAccess.sxSetCooler(coolerInfo.camHandle, coolerStatus, setPoint, coolerStatusPtr, coolerDataPtr);

                // Copy the returnrd data in unmanaged memory back to managed memory
                coolerData = Marshal.ReadInt16(coolerDataPtr);
                coolerStatus = Marshal.ReadByte(coolerStatusPtr, 0);
                if(coolerStatus > 0)
                    coolerInfo.coolerState = true;
                else
                    coolerInfo.coolerState = false;
                coolerInfo.currentTemperature = (Single)((coolerData - 2730)/10.0);
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(coolerDataPtr);
                Marshal.FreeHGlobal(coolerStatusPtr);
            }
            return (coolerInfo);
        }

        // ------------------------------------------------------------
        // Reurns the cooler temperature
        public static Single coolerTemp(IntPtr camHandle)
        {
            Int16 coolerData = 0;   // returned temperature
            Byte coolerStatus = 0;  // assume OFF
            Single currentTemperature = 100;
            Int32 result;

            // Initialize unmanaged memory to hold the struct.
            IntPtr coolerDataPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerData));
            IntPtr coolerStatusPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerStatus));
            try
            {
                result = (Int32)SxSDKAccess.sxGetCoolerTemp(camHandle, coolerStatusPtr, coolerDataPtr);

                // Copy the returned data in unmanaged memory back to managed memory
                coolerData = Marshal.ReadInt16(coolerDataPtr);
                currentTemperature = (Single)((coolerData - 2730) / 10.0);
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(coolerDataPtr);
                Marshal.FreeHGlobal(coolerStatusPtr);
            }
            return (currentTemperature);
        }

        // ------------------------------------------------------------
        // Sets the timer
        public static Int32 setTimer(IntPtr camHandle, UInt32 time)
        {
            return ((Int32)SxSDKAccess.sxSetTimer(camHandle, time));
        }

        // ------------------------------------------------------------
        // Gets the timer value
        public static UInt32 getTimer(IntPtr camHandle)
        {
            return (SxSDKAccess.sxGetTimer(camHandle));
        }

        // --------------------------------------------------------

        // --------------------------------------------------------


    }
}
