﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;
using System.Drawing.Imaging;
using System.IO;

namespace CsdkTester
{
    public partial class Form1 : Form
    {

        const Int32 maxNumberOfCameras = 20;

        IntPtr[] camHandles = new IntPtr[maxNumberOfCameras];
        sxInfo info = new sxInfo();
        public sxInfoPack[] cameraInfo = new sxInfoPack[maxNumberOfCameras];
        sxExposurePack camImgData = new sxExposurePack();
        sxExposurePack[] cameraImageData = new sxExposurePack[maxNumberOfCameras];
        sxModels modelData = new sxModels();
        Int32 numberOfCameras = 0;
        public Int32 selectedCamera = 0;
        Int32 Result = 0;

        UInt16 normalisedTemperature = 2730;
        Boolean setPointChanged = true;
        public Boolean imageFlag = false;
        Boolean imageMemFlag = false;
        Boolean abortExposure = false;
        static Boolean readActive = false;
        UInt32 timerValue = 0;

        Int32 pixelsInImage = 0;
        public UInt16[] pixelBuffer = new UInt16[12000000]; // > largest size image currently available!
        MemoryStream ms = new MemoryStream();       // disposed of in Form Exit
		UInt32 versionDLLmsd, versionDLLlsd, buildDLL;

		// Flags (go in CMD_VALUE_L:H fields)
		// Wipe CCD of all accumulated charge & clear vertical + horizontal registers
		public const UInt16 CCD_WIPE_ALL = 0x00;
		// Specify odd or even field for MX cameras
		public const UInt16 CCD_FLAGS_FIELD_ODD = 0x01;
		// Specify even field for MX cameras
		public const UInt16 CCD_FLAGS_FIELD_EVEN = 0x02;
		// Specify even field for MX cameras
		public const UInt16 CCD_FLAGS_FIELD_BOTH = 0x03;
		// Was no bin accumulate flag
		public const UInt16 CCD_FLAGS_SPARE2 = 0x04;
		// Don't WIPE CDCD when clearing frame
		public const UInt16 CCD_FLAGS_NOWIPE_FRAME = 0x08;
		// Spare flag
		public const UInt16 CCD_FLAGS_SPARE4 = 0x10;
		// Implement TDI (drift scan) operation
		public const UInt16 CCD_FLAGS_TDI = 0x20;
		// Don't clear frame, even when asked
		public const UInt16 CCD_FLAGS_NOCLEAR_FRAME = 0x40;
		// Do not clear vertical & horizontal registers
		public const UInt16 CCD_FLAGS_NOCLEAR_REGISTER = 0x80;
		// Upper bits in byte of CCD_FLAGS word
		public const UInt16 CCD_FLAGS_SPARE8 = 0x01;
		public const UInt16 CCD_FLAGS_SPARE9 = 0x02;
		public const UInt16 CCD_FLAGS_SPARE10 = 0x04;
		public const UInt16 CCD_FLAGS_SPARE11 = 0x08;
		public const UInt16 CCD_FLAGS_SPARE12 = 0x10;
		public const UInt16 CCD_FLAGS_SHUTTER_MANUAL = 0x20;
		public const UInt16 CCD_FLAGS_SHUTTER_OPEN = 0x40;
		public const UInt16 CCD_FLAGS_SHUTTER_CLOSE = 0x80;

        // Properties
        public Int32 imageWidth
        {
            get { return (cameraInfo[selectedCamera].imageWidth); }
        }

        public Int32 imageHeight
        {
            get { return (cameraInfo[selectedCamera].imageHeight); }
        }

        public Boolean imgFlag
        {
            get { return (imageFlag); }
        }


        public Form1()
        {
            InitializeComponent();
        }

        // Initialise the Form
        private void Form1_Load(object sender, EventArgs e)
        {
            buttonShutter.Enabled = false;
            buttonSetTime.Enabled = false;       // allow timer setting once camera selected
            updateTime();
			versionDLLmsd = SxSDKAccess.sxGetDLLVersion();
			buildDLL = versionDLLmsd & 0xFFFF;
			versionDLLmsd = versionDLLmsd >> 16;
			versionDLLlsd = versionDLLmsd & 0xFF;
			versionDLLmsd = versionDLLmsd >> 8;
			label2.Text = "DLL Version " + versionDLLmsd.ToString() + "." + versionDLLlsd.ToString("D2") + ", Build " + buildDLL.ToString("D4");
            for (Int32 index = 0; index < maxNumberOfCameras; index++)
                cameraImageData[index] = camImgData;
//            sxImage.Show();
         }

        // ---- Close any open handles on exit ----
        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if(numberOfCameras > 0)
            {
                for (Int32 index = 0; index >= numberOfCameras; index++)
                {
                    SxSDKAccess.sxClose(camHandles[index]);
                }
            }
            ms.Dispose();   // explicitly close the memory stream
//            sxImage.Dispose();
        }

        // --- Connect Button ----
        // Find out how many cameras are fitted (maximum 20)
        // Retreive information from each of them.
        private void buttonConnect_Click(object sender, EventArgs e)
        {
            textBoxInfo.Text += "SxSDK Tester Utility\r\n";
            numberOfCameras = SxSDKAccess.sxOpen(camHandles);

            if (numberOfCameras > 0)
            {
                buttonConnect.Enabled = false;
                buttonConnect.Text = "Connected";
                textBoxInfo.Text += "Number of cameras = " + numberOfCameras.ToString() + "\r\n\r\n";

                for (Int32 index = 0; index < numberOfCameras; index++)
                {
                    textBoxInfo.Text += "Handle" + index.ToString() + " = " + camHandles[index].ToString() + "\r\n";
                    cameraInfo[index] = sxInfo.readInfoPack(camHandles[index]);//, ref tempInfo);
                    if (cameraInfo[index].result == 1)
                    {
                        // Display some of the data
                        textBoxInfo.Text += "Firmware = " + cameraInfo[index].upperFirmwareVersion.ToString() + "."
                                             + cameraInfo[index].lowerFirmwareVersion.ToString("D2") + "\r\n";
                        textBoxInfo.Text += "Build = " + cameraInfo[index].buildNumber.ToString() + "\r\n";
                        textBoxInfo.Text += "Model = 0x" + cameraInfo[index].cameraModel.ToString("X2")
                                             + ", Name = " + cameraInfo[index].cameraName + "\r\n";
                        textBoxInfo.Text += "Width = " + cameraInfo[index].imageWidth.ToString() + "\r\n";
                        textBoxInfo.Text += "Height = " + cameraInfo[index].imageHeight.ToString() + "\r\n";
                        textBoxInfo.Text += "Fitted = ";
                        if (cameraInfo[index].hasStar2K)
                            textBoxInfo.Text += "Star2K ";
                        if (cameraInfo[index].hasEEProm)
                            textBoxInfo.Text += "EEprom ";
                        if (cameraInfo[index].hasGuider)
                            textBoxInfo.Text += "Guider ";
                        if (cameraInfo[index].hasCooler)
                            textBoxInfo.Text += "Cooler ";
                        if (cameraInfo[index].hasShutter)
                            textBoxInfo.Text += "Shutter ";
                        if (cameraInfo[index].isInterlaced)
                            textBoxInfo.Text += "Interlaced CCD ";
                        textBoxInfo.Text += "\r\n\r\n";
                        comboBoxCamera.Items.Add(cameraInfo[index].cameraName);
                    }
                }
                labelSelect.Visible = true;
                comboBoxCamera.Visible = true;
            }
            else
                textBoxInfo.Text += "No Sx Cameras found!\r\n";
        }

        // Selection of a connected camera
        private void comboBoxCamera_SelectedIndexChanged(object sender, EventArgs e)
        {
            buttonSetTime.Enabled = true;       // allow timer setting once camera selected
            timer1.Enabled = false;             // assume no cooler
            checkBoxCooler.Checked = false;     // assume no cooler
            checkBoxCooler.Enabled = false;     // assume no cooler
            buttonShutter.Enabled = false;      // assume no shutter
            selectedCamera = comboBoxCamera.SelectedIndex;
            textBoxInfo.Text += "selectedCamera = " + selectedCamera.ToString() + "\r\n";
            textBoxInfo.Text += "hasShutter = " + cameraInfo[selectedCamera].hasShutter.ToString() + "\r\n";
            textBoxInfo.Text += "Name = " + cameraInfo[selectedCamera].cameraName + "\r\n";
            if (cameraInfo[selectedCamera].hasShutter)
            {
                buttonShutter.Enabled = true;
				SxSDKAccess.sxSetShutter(camHandles[selectedCamera], 1);
            }
            if (cameraInfo[selectedCamera].hasCooler)
            {
                timer1.Enabled = true;
                checkBoxCooler.Enabled = true;
                checkBoxCooler.Checked = true;
                setPointChanged = true;
            }
        }

        // Use to control the shutter if the camera is equipped with one.
        private void buttonShutter_Click(object sender, EventArgs e)
        {
            Int32 returnData = 0;

            if (buttonShutter.Text == "Close")
            {
				returnData = SxSDKAccess.sxSetShutter(camHandles[selectedCamera], 1);
                buttonShutter.Text = "Open";
            }
            else
            {
				returnData = SxSDKAccess.sxSetShutter(camHandles[selectedCamera], 0);
                buttonShutter.Text = "Close";
            }
            textBoxInfo.Text += "Shutter Data = 0x" + returnData.ToString("X2") + "\r\n";
        }

        // Sets the Cooler temperature
        private void trackBarTemp_ValueChanged(object sender, EventArgs e)
        {
            normalisedTemperature = (UInt16)(trackBarTemp.Value + 2730);
            labelSetTemp.Text = ((Single)((normalisedTemperature - 2730.0) / 10.0)).ToString();
            setPointChanged = true;
        }

        private void checkBoxCooler_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxCooler.Checked)
            {
                checkBoxCooler.Text = "ON";
                setPointChanged = true;         // ensure set point is updated after switch on
            }
            else
                checkBoxCooler.Text = "OFF";
        }

        // Use timer to update cooler temperature
        // Will only update if camera is fitted with cooler and there
        // is no image download pending or in progress
        private void timer1_Tick(object sender, EventArgs e)
        {
            if (!readActive && cameraInfo[selectedCamera].hasCooler)
            {
                if (setPointChanged)
                {
                    setPointChanged = false;
                    sxCoolerPack coolerInfo = new sxCoolerPack();
                    coolerInfo.camHandle = camHandles[selectedCamera];
                    coolerInfo.coolerON = checkBoxCooler.Checked;
                    coolerInfo.setPoint = (Single)((normalisedTemperature - 2730.0) / 10.0);

                    cooler(coolerInfo);
                    Boolean actualState = coolerInfo.coolerState;
                    labelActualTemp.Text = coolerInfo.currentTemperature.ToString();
                }
                else
                    labelActualTemp.Text = coolerTemp(camHandles[selectedCamera]).ToString();

            }
        }

        // Sets exposure time along with  time range Radio buttons
        private void trackBarExposure_ValueChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        // Sets the camera timer (test program)
        private void buttonSetTime_Click(object sender, EventArgs e)
        {
            if ((numberOfCameras > 0) && (timerValue > 0))    // must be connected
            {
                Int32 Result = (Int32)SxSDKAccess.sxSetTimer(camHandles[selectedCamera], timerValue);
                timer2.Enabled = true;
            }
        }

        // Monitors the camera timer when active (test program)
        private void timer2_Tick(object sender, EventArgs e)
        {
            UInt32 time = SxSDKAccess.sxGetTimer(camHandles[selectedCamera]);
            if (time == 0)
            {
                timer2.Enabled = false;
                labelGetTime.Text = "Time remaining = 0 ms";
            }
            else
                labelGetTime.Text = "Time remaining = " + time.ToString() + "ms";
        }

        // Sets the exposure time in timerValue in units of ms using the track bar & radio buttons
        private void updateTime()
        {
            Int32 multiplier = 1000;        // assume milliseconds
            if (radioButtonMinutes.Checked)
            {
                multiplier = 60000;
                labelSetTime.Text = trackBarExposure.Value.ToString() + " mins";
            }
            if (radioButtonSeconds.Checked)
            {
                multiplier = 1000;
                labelSetTime.Text = trackBarExposure.Value.ToString() + " secs";
            }
            if (radioButtonTenths.Checked)
            {
                multiplier = 100;
                labelSetTime.Text = (trackBarExposure.Value * 100).ToString() + " ms";
            }
            if (radioButtonMs.Checked)
            {
                multiplier = 1;
                labelSetTime.Text = (trackBarExposure.Value).ToString() + " ms";
            }
            timerValue = (UInt32)(trackBarExposure.Value * multiplier);
        }

        private void radioButtonMinutes_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        private void radioButtonSeconds_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        private void radioButtonTenths_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        private void radioButtonMs_CheckedChanged(object sender, EventArgs e)
        {
            updateTime();
        }

        // Displays the image on a panel converting from pixel array to 8 bit BMP
        private void panelImage_Paint(object sender, PaintEventArgs e)
        {
            if (imageFlag)
            {
                if (!imageMemFlag)
                {
                    Bitmap imageSxCam = new Bitmap(cameraInfo[selectedCamera].imageWidth, cameraInfo[selectedCamera].imageHeight, PixelFormat.Format8bppIndexed);
                    BitmapData bmpData = imageSxCam.LockBits(new Rectangle(0, 0, imageSxCam.Width, imageSxCam.Height), ImageLockMode.WriteOnly, imageSxCam.PixelFormat);
                    Int32 numBytes = bmpData.Stride * imageSxCam.Height;
                    Int32 i, j, k, index;

                    //Construct the Grayscale colour Palette
                    ColorPalette grayScalePalette = imageSxCam.Palette;
                    for (i = 0; i < 256; i++)
                        grayScalePalette.Entries[i] = Color.FromArgb(255, i, i, i);
                    imageSxCam.Palette = grayScalePalette;

                    // Take msbs of pixels (Int16) array & store in byte array for bmp data
                    Byte[] bmpBuffer = new Byte[numBytes];
                    index = 0;      // index to Int16 pixels in pixelBuffer
                    for (i = 0; i < imageSxCam.Height; i++)
                    {
                        k = i * bmpData.Stride;
                        for (j = 0; j < imageSxCam.Width; j++)
                            bmpBuffer[k++] = (Byte)(pixelBuffer[index++] >> 8);
                    }

                    //Copy the data from the byte array into BitmapData.Scan0
                    Marshal.Copy(bmpBuffer, 0, bmpData.Scan0, numBytes);

                    //Unlock the pixels
                    imageSxCam.UnlockBits(bmpData);

                    // Create Point for upper-left corner of image.
                    Point ulCorner = new Point(0, 0);

                    // Draw image to screen.
                    e.Graphics.DrawImage(imageSxCam, ulCorner);
                    imageSxCam.Dispose();
                    imageFlag = false;
                }
                else
                {
                    Bitmap imageSxCam = new Bitmap(ms);
                    imageMemFlag = false;

                    // Create Point for upper-left corner of image.
                    Point ulCorner = new Point(0, 0);

                    // Draw image to screen.
                    e.Graphics.DrawImage(imageSxCam, ulCorner);
                    imageSxCam.Dispose();
                    imageFlag = false;
                }

            }
        }
        
        // Takes an Image from the selected camera. Returns when the image is downloaded
        private void buttontakeImage_Click(object sender, EventArgs e)
        {
            pixelsInImage = cameraInfo[selectedCamera].imageWidth * cameraInfo[selectedCamera].imageHeight;
            labelPixelNumber.Text = "0";

            cameraImageData[selectedCamera].camHandle = camHandles[selectedCamera];
			cameraImageData[selectedCamera].flags = CCD_FLAGS_FIELD_BOTH; // specify both fields for H series cameras
            cameraImageData[selectedCamera].useGuider = false;
            cameraImageData[selectedCamera].useGated = false;
            cameraImageData[selectedCamera].xOffset = 0;
            cameraImageData[selectedCamera].yOffset = 0;
            cameraImageData[selectedCamera].imageWidth = (UInt16)cameraInfo[selectedCamera].imageWidth;
            cameraImageData[selectedCamera].imageHeight = (UInt16)cameraInfo[selectedCamera].imageHeight;
            cameraImageData[selectedCamera].xBin = 1;
            cameraImageData[selectedCamera].yBin = 1;
            cameraImageData[selectedCamera].exposureTime = timerValue;

            getImage((Int32)timerValue);
            labelPixelNumber.Text = pixelsInImage.ToString();
        }


        // Exposes the image only - must be follwed by 
        private void buttonExpRead_Click(object sender, EventArgs e)
        {
            {
                Boolean waitForExposureEnd = false;
                pixelsInImage = cameraInfo[selectedCamera].imageWidth * cameraInfo[selectedCamera].imageHeight;
                labelPixelNumber.Text = "0";
                if (cameraInfo[selectedCamera].hasShutter)
					SxSDKAccess.sxSetShutter(camHandles[selectedCamera], 1);
                cameraImageData[selectedCamera].camHandle = camHandles[selectedCamera];
				cameraImageData[selectedCamera].flags = CCD_FLAGS_FIELD_BOTH; // specify both fields for H series cameras
                cameraImageData[selectedCamera].useGuider = false;
                cameraImageData[selectedCamera].useGated = false;
                cameraImageData[selectedCamera].xOffset = 0;
                cameraImageData[selectedCamera].yOffset = 0;
                cameraImageData[selectedCamera].imageWidth = (UInt16)cameraInfo[selectedCamera].imageWidth;
                cameraImageData[selectedCamera].imageHeight = (UInt16)cameraInfo[selectedCamera].imageHeight;
                cameraImageData[selectedCamera].xBin = 1;
                cameraImageData[selectedCamera].yBin = 1;
                cameraImageData[selectedCamera].exposureTime = timerValue;
                // For exposures of less than 1 second wait for the exposure to complete
                if (timerValue < 1000)
                    waitForExposureEnd = true;

                if (cameraInfo[selectedCamera].hasShutter)
                    SxSDKAccess.sxSetShutter(camHandles[selectedCamera], 0);;
                Result = makeExposure(cameraImageData[selectedCamera], waitForExposureEnd);
				textBoxInfo.Text += "Result = " + Result.ToString() + ",  Time remaining = " + SxSDKAccess.sxGetTimer(cameraImageData[selectedCamera].camHandle).ToString() + "\r\n";

                // Start the exposure checking timer
                if (!waitForExposureEnd)
                {
                    timerExp.Enabled = true;
                    buttonAbort.Enabled = true;
                }
                else        // Short exposure is already completed so read image now
                {
                    if (cameraInfo[selectedCamera].hasShutter)
						SxSDKAccess.sxSetShutter(camHandles[selectedCamera], 1);
                    getImage(0);
                }
            }
        }

        // This timer is used to monitor the exposure progress,
        // which allows the exposure to be aborted if needed.
        // If the exposure has completed then the image is downloaded.
        // During the image download the "readActive" flag is set
        // This flag should be used to stop any camera communications
        // which requires returned data, e.g. requesting the cooler temperature.
        private void timerExp_Tick(object sender, EventArgs e)
        {
            if (abortExposure == true)
            {
                timerExp.Enabled = false;
                buttonAbort.Enabled = false;
                abortExposure = false;
                labelGetTime.Text = "Time remaining = 0 ms";
                textBoxInfo.Text += "Exposure Aborted" + "\r\n";
            }

            // Get the time left, if zero then exposure completed.
            UInt32 currentTime = SxSDKAccess.sxGetTimer(cameraImageData[selectedCamera].camHandle);
            if (currentTime == 0)
            {
                timerExp.Enabled = false;
                buttonAbort.Enabled = false;
                abortExposure = false;
                labelGetTime.Text = "Time remaining = 0 ms";
                if (cameraInfo[selectedCamera].hasShutter)
					SxSDKAccess.sxSetShutter(camHandles[selectedCamera], 1);
                getImage(0);
            }
            else
            {
                labelGetTime.Text = "Time remaining = " + currentTime.ToString() + "ms";
            }

        }

        private void getImage(Int32 expTime)
        {
            Int32 time;

            readActive = true;
            time = Environment.TickCount & Int32.MaxValue;
            if(expTime == 0)
                Result = readImage(cameraImageData[selectedCamera], ref pixelBuffer);
            else
                Result = makeImage(cameraImageData[selectedCamera], ref pixelBuffer);

            time = (Environment.TickCount & Int32.MaxValue) - time;
            time -= expTime;
            labelDn.Text = time.ToString() + " ms   ";
            readActive = false;
            textBoxInfo.Text += "Read result = " + Result.ToString() + "\r\n";
            labelPixelNumber.Text = pixelsInImage.ToString();
            imageFlag = true;
            panelImage.Invalidate();
        }

        // When the exposure Abort button is pressed set the "abortExposure" flag
        private void buttonAbort_Click(object sender, EventArgs e)
        {
            abortExposure = true;
        }

        // ------------------------------------------------------------
        // Controls the cooler
        private sxCoolerPack cooler(sxCoolerPack coolerInfo)
        {
            Int16 coolerData = 0;   // returned temperature
            UInt16 setPoint = (UInt16)((coolerInfo.setPoint + 273.0) * 10.0); // cooler set point as integer
            Byte coolerStatus = 0; // assume OFF
            if (coolerInfo.coolerON)
                coolerStatus = 1;

            // Initialize unmanaged memory to hold the struct.
            IntPtr coolerDataPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerData));
            IntPtr coolerStatusPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerStatus));
            try
            {
                coolerInfo.result = (Int32)SxSDKAccess.sxSetCooler(coolerInfo.camHandle, coolerStatus, setPoint, coolerStatusPtr, coolerDataPtr);

                // Copy the returnrd data in unmanaged memory back to managed memory
                coolerData = Marshal.ReadInt16(coolerDataPtr);
                coolerStatus = Marshal.ReadByte(coolerStatusPtr, 0);
                if (coolerStatus > 0)
                    coolerInfo.coolerState = true;
                else
                    coolerInfo.coolerState = false;
                coolerInfo.currentTemperature = (Single)((coolerData - 2730) / 10.0);
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(coolerDataPtr);
                Marshal.FreeHGlobal(coolerStatusPtr);
            }
            return (coolerInfo);
        }

        // ------------------------------------------------------------
        // Returns the cooler temperature
        private Single coolerTemp(IntPtr camHandle)
        {
            Int16 coolerData = 0;   // returned temperature
            Byte coolerStatus = 0;  // assume OFF
            Single currentTemperature = 100;
            Int32 result;

            // Initialize unmanaged memory to hold the struct.
            IntPtr coolerDataPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerData));
            IntPtr coolerStatusPtr = Marshal.AllocHGlobal(Marshal.SizeOf(coolerStatus));
            try
            {
                result = (Int32)SxSDKAccess.sxGetCoolerTemp(camHandle, coolerStatusPtr, coolerDataPtr);

                // Copy the returned data in unmanaged memory back to managed memory
                coolerData = Marshal.ReadInt16(coolerDataPtr);
                currentTemperature = (Single)((coolerData - 2730) / 10.0);
            }
            finally
            {
                // Release the unmanaged memory
                Marshal.FreeHGlobal(coolerDataPtr);
                Marshal.FreeHGlobal(coolerStatusPtr);
            }
            return (currentTemperature);
        }

		// Takes exposure & reads pixels from camera when exposure complete
		// using information in sxExposurePack
		private Int32 makeImage(sxExposurePack exposureInfo, ref UInt16[] pixelBuffer)
		{
			Int32 Result = 0;
			Int32 pixelsInImage = (exposureInfo.imageWidth / exposureInfo.xBin) * (exposureInfo.imageHeight / exposureInfo.yBin);
			Int32 pixelA = 0, pixelB = 0, line = 0, linePixel = 0, finishedWidth = 0, finishedHeight = 0, count = 0;
			Int32 cameraModel = getModel(exposureInfo.camHandle);
			Boolean isInterlaced = (cameraModel & 0x8000) > 0;
			cameraModel &= 0xFF;
			UInt16 scale = (UInt16)((cameraModel == sxModels.M25C) ? 2 : 1);		 	 // Adjust width & Height for M25
			UInt16 actualWidth = (UInt16)(exposureInfo.imageWidth * scale);
			UInt16 actualHeight = (UInt16)(exposureInfo.imageHeight / scale);
			UInt16 actualX = (UInt16)(exposureInfo.xOffset * scale);
			UInt16 actualY = (UInt16)(exposureInfo.yOffset / scale);
			finishedHeight = actualHeight * 2;
			finishedWidth = actualWidth / 2;

			Result = SxSDKAccess.sxClearPixels(exposureInfo.camHandle, CCD_WIPE_ALL, (UInt16)(exposureInfo.useGuider ? 1 : 0));

			Result = SxSDKAccess.sxExposePixels(exposureInfo.camHandle, exposureInfo.flags, (UInt16)(exposureInfo.useGuider ? 1 : 0),
													  actualX, actualY, actualWidth, actualHeight,
													  exposureInfo.xBin, exposureInfo.yBin, exposureInfo.exposureTime);

			Int16[] pxBuffer = new Int16[pixelsInImage];
			Int32 bufferSize = Marshal.SizeOf(pxBuffer[0]) * pixelsInImage;
			IntPtr pixelBufferPtr = Marshal.AllocHGlobal(bufferSize);
			try
			{
                Result = SxSDKAccess.sxReadPixels(exposureInfo.camHandle, pixelBufferPtr, (UInt32)(pixelsInImage));
				// Copy the pixelBuffer in unmanaged memory back to managed memory
				Marshal.Copy(pixelBufferPtr, pxBuffer, 0, pixelsInImage);
				System.Buffer.BlockCopy(pxBuffer, 0, pixelBuffer, 0, pixelsInImage * 2);

				// Adjust the Readout for M25C
				if (cameraModel == sxModels.M25C)
				{
					UInt16[] tempBuffer = new UInt16[pixelsInImage];
					for (count = 0; count < pixelsInImage; count++)
						tempBuffer[count] = pixelBuffer[count];

					for (line = 0; line < actualHeight; line++)
					{
						pixelA = line * actualWidth;
						pixelB = pixelA;
						for (linePixel = 0; linePixel < finishedWidth; linePixel++)
						{
							pixelBuffer[pixelA] = tempBuffer[pixelB];
							pixelB++;
							pixelBuffer[pixelA + finishedWidth] = tempBuffer[pixelB];
							pixelB++;
							pixelA++;
						}
					}
				}
			}
			finally
			{
				// Release the unmanaged memory
				Marshal.FreeHGlobal(pixelBufferPtr);
			}
			return (Result);
		}

		// Use this directly after "makeExposure"
		// Reads exposed image immmediately using information in sxExposurePack
		// The exposure time data is ignored
		private Int32 readImage(sxExposurePack exposureInfo, ref UInt16[] pixelBuffer)
		{
			Int32 Result = 0;
			Int32 pixelsInImage = (exposureInfo.imageWidth / exposureInfo.xBin) * (exposureInfo.imageHeight / exposureInfo.yBin);
			Int32 pixelA = 0, pixelB = 0, line = 0, linePixel = 0, finishedWidth = 0, finishedHeight = 0, count = 0;
			Int32 cameraModel = getModel(exposureInfo.camHandle);
			Boolean isInterlaced = (cameraModel & 0x8000) > 0;
			cameraModel &= 0xFF;
			UInt16 scale = (UInt16)((cameraModel == sxModels.M25C) ? 2 : 1);		 	 // Adjust width & Height for M25
			UInt16 actualWidth = (UInt16)(exposureInfo.imageWidth * scale);
			UInt16 actualHeight = (UInt16)(exposureInfo.imageHeight / scale);
			UInt16 actualX = (UInt16)(exposureInfo.xOffset * scale);
			UInt16 actualY = (UInt16)(exposureInfo.yOffset / scale);
			finishedHeight = actualHeight * 2;
			finishedWidth = actualWidth / 2;

			Result = SxSDKAccess.sxClearPixels(exposureInfo.camHandle, CCD_FLAGS_NOWIPE_FRAME, (UInt16)(exposureInfo.useGuider ? 1 : 0));
			Result = SxSDKAccess.sxLatchPixels(exposureInfo.camHandle, exposureInfo.flags, (UInt16)(exposureInfo.useGuider ? 1 : 0),
													  actualX, actualY, actualWidth, actualHeight,
													  exposureInfo.xBin, exposureInfo.yBin);

			Int16[] pxBuffer = new Int16[pixelsInImage];
			Int32 bufferSize = Marshal.SizeOf(pxBuffer[0]) * pixelsInImage;
			IntPtr pixelBufferPtr = Marshal.AllocHGlobal(bufferSize);
			try
			{
                Result = SxSDKAccess.sxReadPixels(exposureInfo.camHandle, pixelBufferPtr, (UInt32)(pixelsInImage));
                // Copy the pixelBuffer in unmanaged memory back to managed memory
				Marshal.Copy(pixelBufferPtr, pxBuffer, 0, pixelsInImage);
				System.Buffer.BlockCopy(pxBuffer, 0, pixelBuffer, 0, pixelsInImage * 2);

				// Adjust the Readout for M25C
				if (cameraModel == sxModels.M25C)
				{
					UInt16[] tempBuffer = new UInt16[pixelsInImage];
					for (count = 0; count < pixelsInImage; count++)
						tempBuffer[count] = pixelBuffer[count];

					for (line = 0; line < actualHeight; line++)
					{
						pixelA = line * actualWidth;
						pixelB = pixelA;
						for (linePixel = 0; linePixel < finishedWidth; linePixel++)
						{
							pixelBuffer[pixelA] = tempBuffer[pixelB];
							pixelB++;
							pixelBuffer[pixelA + finishedWidth] = tempBuffer[pixelB];
							pixelB++;
							pixelA++;
						}
					}
				}
			}
			finally
			{
				// Release the unmanaged memory
				Marshal.FreeHGlobal(pixelBufferPtr);
			}
			return (Result);
		}

		// Takes exposure using information in sxExposurePack
		private Int32 makeExposure(sxExposurePack exposureInfo, Boolean waitForExposureCompleted)
		{
			Int32 Result = 0;

			Result = SxSDKAccess.sxClearPixels(exposureInfo.camHandle, CCD_WIPE_ALL, (UInt16)(exposureInfo.useGuider ? 1 : 0));
			SxSDKAccess.sxSetTimer(exposureInfo.camHandle, exposureInfo.exposureTime);

			if (waitForExposureCompleted)
			{
				while (SxSDKAccess.sxGetTimer(exposureInfo.camHandle) > 0) ;
			}

			return (Result);
		}

		// Determines if the camera is fitted with an interlaced CCD using the camera model number
		// Returns the camera model in bits 0-7 & sets bit15 if the CCD is interlaced.
		private Int32 getModel(IntPtr cameraHandle)
		{
			Int32 cameraModel = (Int32)SxSDKAccess.sxGetCameraModel(cameraHandle);
			Boolean isInterlaced = false;
			// Check for Interlaced CCD in selected camera
			if (((cameraModel & sxModels.MxBit) > 0) || (cameraModel == sxModels.MX5C))
				isInterlaced = true;
			// Although interlaced the M25 has a different readout structure. To deal with this treat the scan as progressive.
			// A special pixel swapping program rearranges them into the correct format for the M25.
			// The H674, H694 & H290 all have the MXbit set erroneously in the Model number, so correct this here.
			if ((cameraModel == sxModels.M25C) || (cameraModel == sxModels.H674) || (cameraModel == sxModels.H694) || (cameraModel == sxModels.H290))
				isInterlaced = false;
			return cameraModel | (isInterlaced ? 0x8000 : 0);
		}


    }
}
